/*--------------------------------------------------------------------------
 *
 *                         -- C N R S --
 *         Laboratoire d'Automatique et d'Analyse des Systemes
 *                 7 Avenue du colonel Roche
 *                     31 077 Toulouse Cedex
 *
 *  Auteur           : Sara Fleury
 *  Groupe           : Robotique et Intelligence Artificielle
 *  Nom du fichier   : piloBokenLineLib.c
 *  Date de creation : 1994
 *
 *  Description      : Lissage de trajectoires type ligne brisse.
 *
 *  Copyright (C) 1994 LAAS/CNRS.
 *----------------------------------------------------------------------*/
#ifdef UNIX
#  include <stdio.h>
#  include <portLib.h>
#else
#  include <vxWorks.h>
#endif
#include <stdio.h>
#include <math.h>
#include <tgmath.h> /* for min() */

/* infinity() is not in all math lib */
static double infinity() {return 1.E10;}

#include "h2mathLib.h"
#include "codels/piloCmdTaskCodels.h"
#include "piloStruct.h"
#include "piloConst.h"
#include "piloBrokenLineLibProto.h"

static void   computeSimpleTurn (int numPt, int seg1, double *longVirage);
static STATUS computeCuspTurn   (int numPt,int numSeg, double *longVirage);

static STATUS zbrent    (double (*func)(double),
                         double a, double b, double tol,
                         double *sol);
static double epsilon_ka(double ka);
static double f_epsilon (double ka);
static double l_ka      (double ka);
static double f_l       (double ka);

/* Variables pour calcul du rebroussement cloth-anti-cloth */
static double _theta;                   /* Variation angulaire robot */
static double _epsilon;                 /* Ecart max a la traj au virage */
static double _l;                       /* Longueur virage */

#define THETA2     (fabs(_theta/2))
#define THETA0(ka) (PILO_MOT_AXIS_WIDTH/(4.0*(ka)))
#define KC(ka)     (sqrt(PI*PILO_MOT_AXIS_WIDTH/2)*(THETA2*sqrt((ka)) - PILO_MOT_AXIS_WIDTH/(4*sqrt(ka))))


/*****************************************************************************
 *
 *  piloSmoothFiltre -  Filtre les donnees d'entre
 *
 *  Description -
 *    Supprime les distances et les angles negligeables.
 *    A l'issu du filtrage:
 *        . il n'y a plus de distance nulle
 *        . il peut subsister des angles nules dans le cas de rebroussement
 *    Le nombre de mouvements elementaires nbMvts retourne est un majorant.
 *    En effet on suppose ici qu'un virage se compose de
 *     CLOTHO[-ANTICLOTHO]-CLOTHO, or cette chaine peut se reduire en une
 *     simple rotation...
 *
 *  Retourne: un majorant du nombre de mouvements elementaires.
 */

STATUS piloSmoothFiltre (int *nbMvts)
{
  int i, j, nb;
  double angleSigne;

  /*
   * Filtrage des petites distances
   */
  j = 0;
  angleSigne = 0;
  nb = DATA_NB;
  for (i = 0; i < nb; i++) {

    angleSigne += DATA_DTH (i);

    /* Distance assez grande ou dernier point */
    if ( fabs(DATA_DL (i)) > EPSILON_DIST || i == nb-1)  {
      DATA_DTH (j) = angleSigne;
      DATA_DL  (j) = DATA_DL (i);
      DATA_EPS (j) = DATA_EPS (i);
      angleSigne = 0;
      j++;
    }
  }
  nb = j;

  /*
   * Filtrage des petits angles sans point de rebroussement
   */

  /* Limitation des angles entre +- 2PI*/
  for (i = 0; i < nb; i++) {
    while (DATA_DTH (i) >  2.*PI) DATA_DTH (i) -= 2.*PI;
    while (DATA_DTH (i) < -2.*PI) DATA_DTH (i) += 2.*PI;
  }
  j = 1;

  for (i = 1; i < nb; i++){

    /* Si petit angle avec faible decalage laterale finale */
    if ( (SIGN(DATA_DL (i)) == SIGN(DATA_DL (i-1)))
        && (fabs(DATA_DTH (i)) < EPSILON_ANGLE)
        && (fabs(DATA_DTH (i) * DATA_DL (i)) < DATA_EPS(i)))

      /* Ajouter la distance suivante a la distance precedente */
      DATA_DL (j-1) += DATA_DL (i);

    /* Sinon enregistrer la consigne et passer a la suivante */
    else {
      DATA_DL  (j) = DATA_DL  (i);
      DATA_DTH (j) = DATA_DTH (i);
      DATA_EPS (j) = DATA_EPS (i);
      j++;
    }
  }
  nb = j;
  /* Enregistrement du nombre de points */
/*  printf ("piloSmoothFiltre: %d points -> %d points\n", DATA_NB, nb);*/
  DATA_NB = nb;

  /*
   * Determination (majorante) du nb de mvts e'lementaires : turn + seg + ...
   */
  if (fabs(DATA_DTH(0)) < EPSILON_ANGLE)
    *nbMvts = 1;
  else
    *nbMvts = 2;
  for (i = 1; i < nb; i++) {

    /* Le dernier virage n'est qu'un spin */
    if ((i==nb-1) && (fabs(DATA_DL (i)) < EPSILON_DIST))
      (*nbMvts) += 1;

    /* Virage non lisse : turn + seg */
    else if (DATA_EPS(i-1) < EPSILON_EPS)
      (*nbMvts) += 2;

    /* Virage sans rebroussement : cloth + cloth + seg */
    else if (SIGN(DATA_DL (i-1)) == SIGN(DATA_DL (i)))
      (*nbMvts) += 3;

    /* Virage avec rebroussement : cloth + anti + cloth + seg */
    else
      (*nbMvts) += 4;
  }
  /* Position finale */
  (*nbMvts)++;
  return(OK);
}


/******************************************************************************
 *
 *  piloSmoothTypes  -  Determination des mouvements elementaires.
 *  piloSmoothGeo  -  Determination de la geometrie des mouvements elementaires.
 *
 *  Description:
 *  La ligne brise'e est transforme'e en une succession de segments de droites
 *  et de virages constitue's de deux demi clotho symetriques lorsqu'il n'y a
 *  pas de rebroussement ou d'une succession cloth-anticlotho-cloth lorsqu'il
 *  y a rebroussement.
 *  Ces virages sont determine's de facon a` respecter:
 *    a) L'ecart max a` la trajectoire (atteint, eventuellement,
 *       au sommet Z du virage);
 *    b) Le virage ne doit pas entamer les segments entre 2 consignes
 *       successives de plus de la longueur de ces segments divise'e par un
 *       facteur FACT_MINOR_L_CLOTH (= 2).
 *  la trajectoire peut debuter et finir par une rotation sur place (spin).
 *
 *  Une fois la geometrie de la trajectoire fixe'e, on determinera les
 *  vitesses et acceleration de parcours, limite'es par les valeurs de
 *  consignes.
 *
 *                     sans rebroussement          avec rebroussement
 *        numMvt :         segment                     segment
 *        numMvt+1 :       clotho                      clotho
 *        numMvt+2 :       clotho                      anti
 *        numMvt+3 :                                   clotho
 *
 *
 *  Retourne:  OK ou bilan
 */

STATUS piloSmoothTypes (int *nbMvts)

{
  int i, numMvt;
  int nbMvtsReal;                        /* Nombre de mouvements recalcule */

  /* Rotation et segment initiaux */
  numMvt = 0;
  i = 0;
  if (fabs(DATA_DTH(i)) >= EPSILON_ANGLE) {
    SEG_NUM(numMvt) = 0;
    TYPE_MVT(numMvt++) = TURN;
  }
  if (fabs(DATA_DL(i)) >= EPSILON_DIST) {
    SEG_NUM(numMvt) = 0;
    TYPE_MVT(numMvt++) = SEGMENT;
  }

  /* Les virages */
  for (i = 1; i < DATA_NB; i++) {

    /* Il reste un virage */
    if ((i != DATA_NB-1) || (fabs(DATA_DL (DATA_NB-1)) > EPSILON_DIST)) {

      /*
       * Virages degeneres
       */
      /* angle nule (avec rebroussement) */
      if (fabs(DATA_DTH (i)) < EPSILON_ANGLE) {
      }
      /* Angle de 2*PI avec rebroussement */
      else if ((SIGN(DATA_DL (i-1)) != SIGN(DATA_DL (i)))
           && (fabs(DATA_DTH (i)) > 2.*PI - EPSILON_ANGLE)) {
        SEG_NUM(numMvt) = i;
        TYPE_MVT(numMvt++) = TURN;
      }
      /* Angle de PI sans rebrousement */
      else if ((SIGN(DATA_DL (i-1)) == SIGN(DATA_DL (i)))
           && (fabs(DATA_DTH (i)) > PI - EPSILON_ANGLE)
           && (fabs(DATA_DTH (i)) < PI + EPSILON_ANGLE)) {
        SEG_NUM(numMvt) = i;
        TYPE_MVT(numMvt++) = TURN;
      }
      /* Ecart a la trajectoire nule */
      else if (fabs(DATA_EPS (i-1) < EPSILON_EPS)) {
        SEG_NUM(numMvt) = i;
        TYPE_MVT(numMvt++) = TURN;
      }

      /*
       * Virage normal
       */
      else {
        SEG_NUM(numMvt) = i-1;
        TYPE_MVT (numMvt++) = CLOTHO;
        /* Rebroussement */
        if (SIGN(DATA_DL (i-1)) != SIGN(DATA_DL (i))) {
          SEG_NUM(numMvt) = i;
          TYPE_MVT (numMvt++) = ANTI_CLOTHO;
        }
        SEG_NUM(numMvt) = i;
        TYPE_MVT (numMvt++) = CLOTHO;
      }

      /*
       * Segment suivant
       */
      SEG_NUM(numMvt) = i;
      TYPE_MVT(numMvt++) = SEGMENT;
    }

    /* On fini sur une rotation pure */
    else {
      SEG_NUM(numMvt) = i;
      TYPE_MVT (numMvt++) = TURN;
    }
  } /* Les virages */

  /* Marque la fin de la traj */
  SEG_NUM(numMvt) = DATA_NB-1;
  TYPE_MVT (numMvt) = GOTO_END;
  nbMvtsReal = numMvt+1;
  if (nbMvtsReal > *nbMvts) {
    printf ("piloSmoothGeo   : %d mvts   -> %d mvts\n", *nbMvts, nbMvtsReal);
    printf ("piloSmoothGeo: incoherence nombre mouvement elementaires\n");
    return ERROR;
  }
  *nbMvts = nbMvtsReal;
  return OK;
}

/******************************************************************************
 *
 *  piloSmoothGeo  -  Determination de la geometrie des mouvements
 *                    elementaires definis par la procedure
 *                    piloSmoothTypes.
 *
 *  Description:
 */

STATUS piloSmoothGeo (int *nbMvts)
{

  int numPt;                             /* Numero pt consigne courant */
  int increment;                         /* Nb de mvts pour le virage */
  int seg1, cloth1, anti, cloth2, seg2, turn;  /* Num mvt elementaire */
  double longVirage;                     /* Longueur prise pour le virage */
  double prevLongVirage;                 /* Precedente longueur longVirage */
  double pVCloth;                       /* Parametre de clothoide au virage */
  double sinPVCloth;                     /* Sin de fresnel de pVCloth */
  double cosPVCloth;                     /* Cos de fresnel de pVCloth */
  double xSommet, ySommet, thetaSommet;  /* Coord globales des sommets de la
                                            ligne brise'e */
  double xCloth, yCloth;                 /* Coord dans repere clotho
                                            du virage */
  int i, numMvt;
  double chs;                            /* Flag changement de sens */
  int nbVirages;
  int nbMvtsReal;                        /* Nombre de mouvements recalcule */


  /* Initialisations */
  prevLongVirage = 0.;
  numMvt = 0;
  numPt = 0;
  thetaSommet = THETA_START(numMvt);
  xSommet = X_START(numMvt);
  ySommet = Y_START(numMvt);

  /* Spin initial */
  if (TYPE_MVT(numMvt) == TURN) {
    RO_START(numMvt) = 0.;
    VAR_ANG(numMvt) = DATA_DTH (numPt);
    thetaSommet += VAR_ANG(numPt);
    numMvt++;
  }

  /* 1er segment (sinon fin) */
  if (TYPE_MVT(numMvt) == SEGMENT) {
    X_START(numMvt) = xSommet;
    Y_START(numMvt) = ySommet;
    THETA_START(numMvt) = thetaSommet;
  }

  /* Calcul du nombre de virages */
  nbVirages = TYPE_MVT(*nbMvts-2) == SEGMENT ? DATA_NB : DATA_NB-1;

  /* Pour chaque virage numPt calcul des mvts elementaires */
  for (numPt = 1; numPt < nbVirages; numPt++) {

    /* Coordonnees globales du point de consigne suivant */
    xSommet += DATA_DL (numPt-1) * cos(thetaSommet);
    ySommet += DATA_DL (numPt-1) * sin(thetaSommet);

    seg1 = numMvt;

    /* Calcul du virage selon le type */
    switch(TYPE_MVT(numMvt+1)) {

    /*
     * Deux segments d'affiles (ie rebroussement externe avec angle nule)
     */
    case SEGMENT:
      seg2 = seg1+1;
      increment = 1;

      /* Longueur de ce virage */
      /*XXXX Ceci est faux, pour trouver la vraie longueur il faut
        la calculer dans le cas general avec l'angle tendant vers 0 */
      longVirage = - SIGN(DATA_DL(numPt-1)) *
        MIN(DATA_EPS(numPt-1),
             MIN(fabs(DATA_DL(numPt-1)), fabs(DATA_DL(numPt)))
             /FACT_MINOR_L_CLOTH/2.);

      /* Longueur de la ligne droite avt virage */
      VAR_LIN (seg1) = DATA_DL (numPt-1) - (prevLongVirage + longVirage);
      VAR_ANG  (seg1) = 0.;

      /* Apres un rebroussement longVirage change de signe */
      longVirage *= -1;
      prevLongVirage = longVirage;

      /* On tourne (meme si petit angle !) */
      thetaSommet += DATA_DTH (numPt);

      /* Debut segment suivant */
      X_START (seg2) = xSommet + longVirage * cos(thetaSommet);
      Y_START (seg2) = ySommet + longVirage * sin(thetaSommet);
      THETA_START (seg2) = THETA_START (seg1);

      break;   /* SEGMENT */

    /*
     * Pure rotation (DATA_EPS = 0 ou demi-tour)
     */
    case TURN:
      turn = seg1+1;
      seg2 = seg1+2;
      increment = 2;

      /* Longueur de ce virage */
      /* Loop */
      if (fabs(DATA_DTH(numPt)) > PI && fabs(DATA_DTH(numPt)) < 3.*PI/2.)
        longVirage = - SIGN(DATA_DL(numPt-1)) * DATA_EPS(numPt-1);
      /* Turn ou rebroussement interne */
      else
        longVirage = SIGN(DATA_DL(numPt-1)) *
          MIN(DATA_EPS(numPt-1),
               MIN(fabs(DATA_DL(numPt-1)), fabs(DATA_DL(numPt)))
               /FACT_MINOR_L_CLOTH);

      /* Longueur de la ligne droite avt virage */
      VAR_LIN (seg1) = DATA_DL (numPt-1) - (prevLongVirage + longVirage);
      VAR_ANG  (seg1) = 0.;

      /* Rotation */
      X_START (turn) = xSommet - longVirage * cos(thetaSommet);
      Y_START (turn) = ySommet - longVirage * sin(thetaSommet);
      THETA_START (turn) = THETA_START (seg1);
      VAR_ANG (turn) = DATA_DTH (numPt);
      VAR_LIN (turn) = 0.;

      /* On tourne */
      thetaSommet += DATA_DTH (numPt);

      /* Apres un rebroussement longVirage change de signe */
      if (SIGN(DATA_DL(numPt-1)) != SIGN(DATA_DL(numPt)))
        longVirage *= -1.;
      prevLongVirage = longVirage;

      /* Debut segment suivant */
      X_START (seg2) = xSommet + longVirage * cos(thetaSommet);
      Y_START (seg2) = ySommet + longVirage * sin(thetaSommet);
      THETA_START (seg2) = THETA_START (turn) + VAR_ANG (turn);

      break;  /* TURN */

    /*
     * Virage lisse
     */
    case CLOTHO:

      /* Caracteristiques virage sans rebroussement: K_TRAJ, VAR_LIN et
         VAR_ANG de cloth1 et cloth2 */
      if (TYPE_MVT(seg1+2) != ANTI_CLOTHO) {
        computeSimpleTurn (numPt, seg1, &longVirage);
        cloth1 = seg1 + 1;
        cloth2 = seg1 + 2;
        seg2   = seg1 + 3;
        increment = 3;
        chs = 1.;
      }

      /* Caracteristiques virage avec rebroussement !!!! externe !!!!! :
         K_TRAJ, VAR_LIN et VAR_ANG de cloth1, anti et cloth2 */
      else {
        if (computeCuspTurn (numPt, seg1, &longVirage) == ERROR)
          return ERROR;
        cloth1 = seg1 + 1;
        anti   = seg1 + 2;
        cloth2 = seg1 + 3;
        seg2   = seg1 + 4;
        increment = 4;
        chs = -1.;
      }

      /* Angles absolus de cloth1, anti eventuellement, cloth2 et seg2 */
      THETA_START(cloth1) = THETA_START(seg1);
      for (i=2; i<=increment; i++) {
        THETA_START(seg1+i) = THETA_START(seg1+i-1) + VAR_ANG(seg1+i-1);
      }

      /* Longueur de la ligne droite avt virage */
      VAR_LIN (seg1) = DATA_DL (numPt-1) - (prevLongVirage + longVirage);
      VAR_ANG  (seg1) = 0.;

      /* Calcul des coordonnees du point de debut de virage */
      X_START (cloth1) = xSommet - longVirage * cos(thetaSommet);
      Y_START (cloth1) = ySommet - longVirage * sin(thetaSommet);

      /* Coordonnees fin clotho dans le repere de la clotho */
      pVCloth = sqrt (2.*fabs(VAR_ANG (seg1+1)) / PI );
      sinPVCloth = integSinFresnel (pVCloth);
      cosPVCloth = integCosFresnel (pVCloth);
      xCloth = K_TRAJ (cloth1) * cosPVCloth;
      yCloth = SIGN(VAR_ANG(cloth1)) * K_TRAJ (cloth1) * sinPVCloth;

      /* Calcul des coordonnees fin cloth1 dans le repere global */
      X_START (cloth1+1) = X_START (cloth1)
        + xCloth * cos(thetaSommet) - yCloth * sin(thetaSommet);
      Y_START (cloth1+1) = Y_START (cloth1)
        + xCloth * sin(thetaSommet) + yCloth * cos(thetaSommet);

      /* On tourne */
      thetaSommet += DATA_DTH (numPt);

      /* Apres un rebroussement longVirage change de signe */
      longVirage *= chs;
      prevLongVirage = longVirage;

      /* Debut segment suivant */
      X_START (seg2) = xSommet + longVirage * cos(thetaSommet);
      Y_START (seg2) = ySommet + longVirage * sin(thetaSommet);

      /* Coord fin d'anti (= debut clotho2) */
      if (increment == 4) {
        X_START(cloth2) = X_START (seg2)
          + xCloth * cos(thetaSommet) + yCloth * sin(thetaSommet);
        Y_START(cloth2) = Y_START (seg2)
          + xCloth * sin(thetaSommet) - yCloth * cos(thetaSommet);
      }
      break;                 /* CLOTHO */

    default:
      increment = 0;
      printf ("piloSmoothGeo  : TYPE_MVT(%d) = %d impossible\n",
              seg1+1, TYPE_MVT(seg1+1));
    }  /* Switch */

    /* Next */
    numMvt += increment;

  }  /* Fin boucle for */

  /* Variation dernier segment */
  if (TYPE_MVT(numMvt) == SEGMENT) {
    VAR_LIN(numMvt) = DATA_DL (numPt-1) - prevLongVirage;
    xSommet += DATA_DL (numPt-1) * cos(thetaSommet);
    ySommet += DATA_DL (numPt-1) * sin(thetaSommet);
    numMvt++;
  }

  /* Spin final */
  if (TYPE_MVT(numMvt) == TURN) {
    X_START(numMvt) = xSommet;
    Y_START(numMvt) = ySommet;
    THETA_START(numMvt) = thetaSommet;
    RO_START(numMvt) = 0.;
    VAR_ANG(numMvt) = DATA_DTH (numPt);
    numMvt++;

    /*  Position finale */
    THETA_START (numMvt) = thetaSommet + VAR_ANG(numMvt-1);
    X_START (numMvt) = xSommet;
    Y_START (numMvt) = ySommet;
  }
  else {
    /* Position final */
    THETA_START (numMvt) = thetaSommet;
    X_START (numMvt) = xSommet;
    Y_START (numMvt) = ySommet;
  }
  if (TYPE_MVT(numMvt) != GOTO_END)
    printf ("piloSmoothGeo: dernier mvt (%d) != GOTO_END\n", numMvt);

  nbMvtsReal = numMvt+1;
  if (nbMvtsReal > *nbMvts) {
    return ERROR;
  }
  *nbMvts = nbMvtsReal;

  return(OK);
}



/*****************************************************************************
 *
 *  computeSimpleTurn  -  Calcul les caracteristiques d'un virage sans
 *                       rebroussement
 *
 *
 *  Description: Calcul les caracteristiques du segment et des deux
 *               demi-clothos symetriques constituant le virage.
 *
 *  Retourne:  rien
 */

static void computeSimpleTurn (int numPt,     /* Numero de point de consigne */
                              int seg1,      /* Numero du mouvement */
                              double *longVirage)

{
  double pVCloth;                       /* Parametre de clothoide au virage */
  double sinPVCloth;                    /* Sin de fresnel de pVCloth */
  double cosPVCloth;                    /* Cos de fresnel de pVCloth */
  double kVirage;                       /* Constante du virage */
  double dTheta;

  /* Variations angulaires */
  dTheta = DATA_DTH (numPt) / 2.;
  VAR_ANG (seg1+1) = dTheta;
  VAR_ANG (seg1+2) = dTheta;
  dTheta = fabs (dTheta);

  /* Calcul du parametre de la clotho au sommet du virage */
  pVCloth = sqrt (fabs(DATA_DTH (numPt)) / PI );

  /* Calcul sin et cos de fresnel normalise' au sommet du virage */
  sinPVCloth = integSinFresnel (pVCloth);
  cosPVCloth = integCosFresnel (pVCloth);

  /* Calcul de la constante de virage */
  kVirage = SIGN(cos (dTheta)) *
    (cos (dTheta) * cosPVCloth + sin (dTheta) * sinPVCloth);

  /* Longueur max du virage */
  *longVirage = MIN( fabs(DATA_DL (numPt-1)), fabs(DATA_DL (numPt)))
                / FACT_MINOR_L_CLOTH ;

  /* Calculer la longueur du virage */
  if (sinPVCloth > 0.0000001)
    *longVirage = MIN(*longVirage, DATA_EPS (numPt-1) * kVirage / sinPVCloth);

  /* Signe */
  *longVirage *= SIGN(DATA_DL (numPt-1));

  /* Calcul de la constante geometrique des clothoides */
  if (fabs(*longVirage) < 0.00000001)
     K_TRAJ (seg1+1) = 0.;
  else {
    if (kVirage > 0.000001)
      K_TRAJ (seg1+1) = SIGN(DATA_DL (numPt)) *
        fabs (cos (dTheta) * *longVirage / kVirage);
    else if (sinPVCloth > 0.0000001)
      K_TRAJ (seg1+1) = SIGN(DATA_DL (numPt)) *
        fabs(cos (dTheta) * DATA_EPS (numPt-1) / sinPVCloth);
    else
      K_TRAJ (seg1+1) = infinity();
  }

  K_TRAJ (seg1+2) = K_TRAJ (seg1+1);

  /* Rayons de gyration initiaux */
  RO_START(seg1+1) = infinity();
  C_START(seg1+1) = 0.;
  RO_START(seg1+2) = SIGN(VAR_ANG (seg1+2))
    * K_TRAJ(seg1+2)/sqrt(2*PI*dTheta);
  C_START(seg1+2) = SIGN(VAR_ANG (seg1+2))
    * sqrt(2*PI*dTheta)/K_TRAJ(seg1+2);

  /* Variation curviligne */
  VAR_LIN (seg1+1) = sqrt(2. * fabs(VAR_ANG(seg1+1)/PI)) * K_TRAJ(seg1+1);
  VAR_LIN (seg1+2) = VAR_LIN (seg1+1);
}


/*****************************************************************************
*
*  computeCuspTurn  -  Calcul les caracteristiques d'un virage avec
*                       rebroussement
*
*
*  Description: Calcul les caracteristiques du segment et des deux demi-clotho
*               symetriques et de l'anti-clotho constituant le virage.
*
*  Retourne:  rien
*/

static STATUS computeCuspTurn (int numPt,      /* Numero de point de consigne */
                              int numSeg,           /* Numero du mouvement */
                              double *longVirage)   /* Longueur du virage */
{
  double l;
  double errabs=1.0e-4;
  double a, b;
  static double ka = 0;
  double l1, l2;
  int cloth1, anti, cloth2;


  cloth1 = numSeg+1;
  anti = numSeg+2;
  cloth2 = numSeg+3;

  /* Definition de variables globales pour les resolutions */
  _theta = DATA_DTH(numPt);
  _epsilon = DATA_EPS(numPt-1);

  /* calcul de ka0 tel que epsilon_ka(ka) - epsilon == 0 */
  a = PILO_MOT_AXIS_WIDTH / (4. * MIN(fabs(_theta), PI)) + 0.0001;
  b = 1.0e8;
  if (zbrent(f_epsilon, a, b, errabs, &ka) == ERROR)
    return ERROR;

  /* calcul de l_ka(ka0) */
  l = l_ka(ka);

  /* limitation de l */
  l1 = fabs(DATA_DL(numPt-1))/FACT_MINOR_L_CLOTH;
  l2 = fabs(DATA_DL(numPt))/FACT_MINOR_L_CLOTH;
  if ((l > l1) || (l > l2)) {

    _l = l = MIN(l1, l2);

    /* recherche de ka tel que l_ka(ka) == l */
    if (zbrent(f_l, a, b, errabs, &ka) == ERROR)
      return ERROR;
  }

  *longVirage = SIGN(DATA_DL(numPt-1)) * l;
  K_TRAJ(anti) = -  SIGN(DATA_DL(numPt-1)) * ka;
  K_TRAJ(cloth1) = SIGN(DATA_DL(numPt-1)) * KC(ka);
  K_TRAJ(cloth2) = -K_TRAJ(cloth1);

  /* Variations angulaires */
  VAR_ANG (cloth1) = SIGN(_theta) * THETA0(ka);
  VAR_ANG (cloth2) = VAR_ANG (cloth1);
  VAR_ANG (anti) = _theta - 2. * VAR_ANG (cloth1);

  /* Variation des distances */
  VAR_LIN(cloth1) = sqrt(2. * fabs(VAR_ANG(cloth1)/PI)) * K_TRAJ(cloth1);
  VAR_LIN(cloth2) = - VAR_LIN(cloth1);
  /* Longueur anti-cloth avec signe du depart de l'anti-cloth */
  VAR_LIN(anti) = - K_TRAJ(anti) * SQR(VAR_ANG (anti)) / 2.;

  /* Rayons de gyration initiaux */
  RO_START(cloth1) = infinity();
  C_START(cloth1)  = 0.;
  if (fabs(VAR_ANG (cloth1)) > 0.000000001)
    RO_START(anti) = SIGN(VAR_ANG (cloth1))
      * K_TRAJ(cloth1)/sqrt(2*PI*fabs(VAR_ANG (cloth1)));
  else
    RO_START(anti) = infinity();

  /* Curbures initiales */
  if (fabs(K_TRAJ(cloth1)) > 0.00000001)
    C_START(anti)    = SIGN(VAR_ANG (cloth1))
      * sqrt(2*PI*fabs(VAR_ANG (cloth1)))/K_TRAJ(cloth1);
  else
     C_START(anti) = infinity();

  RO_START(cloth2) = - RO_START(anti);
  C_START(cloth2)  = - C_START(anti);
  return OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Rebroussements par anti-clothoides
 **/

/*
 * fonctions d'aide
 */

/***************************************************************************/
/* Ecart max a la traj au virage en fonction du parametre de l'anti-clotho */
/***************************************************************************/

static double epsilon_ka(double ka)
{
  double t1, t2, t3;

  t1 = sin(THETA0(ka)) - sin(angleLimit(THETA2))
    + cos(THETA0(ka))*(THETA2 - THETA0(ka));

  t2 = KC(ka)*integSinFresnel(sqrt(PILO_MOT_AXIS_WIDTH/(2*PI*ka)));

  t3 = (ka*t1 + t2)/fabs(sin(_theta/2));

  /* printf("epsilon_ka(%f) = %f\n", ka, t3); */

  return(t3);
}


static double f_epsilon(double ka)
{
    return(epsilon_ka(ka) - _epsilon);
}

static double l_ka(double ka)
{
    double t1, t2, t3;

    t1 = - epsilon_ka(ka)*cos(_theta/2);
    t2 = cos(THETA0(ka)) - cos(THETA2)
        - (THETA2 - THETA0(ka))*sin(THETA0(ka));
    t3 = KC(ka)*integCosFresnel(sqrt(PILO_MOT_AXIS_WIDTH/(2*PI*ka)));

    /* printf("l_ka(%f) = %f\n", ka, t1 + ka*t2 + t3); */
    return(t1 + ka*t2 + t3);
}

static double f_l(double ka)
{
    return(l_ka(ka) - _l);
}


/*----------------------------------------------------------------------*/

#define ITMAX 100
#define EPS 3.0e-8

static void nrerror(char *msg)
{
    printf("zbrent: Numerical Recipes run-time error...\n");
    printf("%s\n",msg);
}

/********************************************************/
/* Fonction de recherche de zero recuperee par Matthieu */
/********************************************************/


static STATUS zbrent(func,x1,x2,tol,sol)
  double (*func)();       /* ANSI: double (*func)(double); */
  double x1,x2,tol;       /* Bornes de recherches + tolerance */
  double *sol;            /* resultat */
{
    int iter;
    double a=x1,b=x2,c=0.,d=0.,e=0.,min1,min2;
    double fa=(*func)(a),fb=(*func)(b),fc,p,q,r,s,tol1,xm;
    void nrerror();

    if (fb*fa > 0.0) {
      nrerror("Root must be bracketed in ZBRENT");
      return ERROR;
    }
    fc=fb;
    for (iter=1;iter<=ITMAX;iter++) {
        if (fb*fc > 0.0) {
            c=a;
            fc=fa;
            e=d=b-a;
        }
        if (fabs(fc) < fabs(fb)) {
            a=b;
            b=c;
            c=a;
            fa=fb;
            fb=fc;
            fc=fa;
        }
        tol1=2.0*EPS*fabs(b)+0.5*tol;
        xm=0.5*(c-b);
        if (fabs(xm) <= tol1 || fb == 0.0) {
          *sol = b;
          return OK;
        }
        if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
            s=fb/fa;
            if (a == c) {
                p=2.0*xm*s;
                q=1.0-s;
            } else {
                q=fa/fc;
                r=fb/fc;
                p=s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
                q=(q-1.0)*(r-1.0)*(s-1.0);
            }
            if (p > 0.0)  q = -q;
            p=fabs(p);
            min1=3.0*xm*q-fabs(tol1*q);
            min2=fabs(e*q);
            if (2.0*p < (min1 < min2 ? min1 : min2)) {
                e=d;
                d=p/q;
            } else {
                d=xm;
                e=d;
            }
        } else {
            d=xm;
            e=d;
        }
        a=b;
        fa=fb;
        if (fabs(d) > tol1)
            b += d;
        else
            b += (xm > 0.0 ? fabs(tol1) : -fabs(tol1));
        fb=(*func)(b);
    }
    nrerror("Maximum number of iterations exceeded in ZBRENT");
    return(ERROR);
}

#undef ITMAX
#undef EPS

