/*- -*- Mode: C -*- ----------------------------------------------------
 *    
 *                         -- C N R S -- 
 *         Laboratoire d'Automatique et d'Analyse des Systemes 
 *                 7 Avenue du colonel Roche   
 *                     31 077 Toulouse Cedex
 *
 *  Auteur           : Sara Fleury
 *  Groupe           : Robotique et Intelligence Artificielle
 *  Nom du fichier   : piloRSLib.c
 *  Date de creation : 1994
 * 
 *  Description      : Lissage de trajectoires type Reed&Shep 
 *
 *
 *  Copyright (C) 1994 LAAS/CNRS.
 *----------------------------------------------------------------------*/
#ifdef UNIX
#  include <stdio.h>
#  include <portLib.h>

/* #  include <sunmath.h> */
#else
#  include <vxWorks.h>
#endif
#include <stdio.h>
#include <math.h>
#include <tgmath.h>

/* infinity() is not in all math lib */
static double infinity() {return 1.E10;}

#include "h2mathLib.h"

#include "piloStruct.h"
#include "piloConst.h"
#include "codels/piloCmdTaskCodels.h"
#include "server/piloError.h"
#include "piloRSLibProto.h"

typedef struct CIRCLE {
  double x, y, r;
} CIRCLE;

typedef enum PILO_CIRCLE_TYPE {
  CIRCLE_START, CIRCLE_CENTER, CIRCLE_END
} PILO_CIRCLE_TYPE;

typedef enum TYPE_MOTION {
  LINE, STOP
}TYPE_MOTION;

#define IS_CIRCLE(th)          ((th) == 0. ? FALSE : TRUE)
#define IS_STOP(sign1sign2)    ((sign1sign2) > 0. ? FALSE : TRUE)
#define PILO_REDUCE_RO 0.95
#define PILO_REDUCE_ANGLE (2./5.)
#define PILO_REDUCE_LENGTH (2./5.)
#define PHI_WANTED          (SQR(WMAX_PILO)/2.*AMAX_ANG_PILO)
#define SQRT_PI             (sqrt(PI))
#define PILO_REDUCED_PHI_J  (2./3.)


static BOOL computeTangentWithPrev(int numData, 
				   PILO_CIRCLE_TYPE prevCircleType, 
				   PILO_CIRCLE_TYPE circleType, 
				   double newRo1, double newRo2,
				   double *dTheta, double *length);
static CIRCLE tangentCircle(CIRCLE *circle, double theta, double newRo);
static STATUS tangentSegBetweenCircles(CIRCLE *c1, CIRCLE *c2, 
				       double *l, double *theta);
static int computeSmoothLineCircle(double lMax, double phiTan, 
				   double phiMax, double ro,
				   double *lV, double *phiV,
				   double *rV, double *kClotho);


/******************************************************************************
 *
 *  piloRSFiltre  -  Calcul du nombre de mouvements elementaires,
 *                   test de la validite des enchainement.
 *
 *  Description: La fonction piloRSgeo ne sait lisser (par une clotho) que les 
 * enchainements droite/cercle. Les enchainements cercles/cercles
 * - sans rebroussement- vont donc etre substitues par des enchainement 
 * cercle/droite/cercle (voir fonction piloRSTypes).
 *
 *  Retourne:  OK ou bilan
 */
int piloRSFiltre (int *nMvts)
{
  BOOL prevIsCircle, isCircle;
  BOOL stopBefore;
  int nbMvts;
  int i, j;

  prevIsCircle = IS_CIRCLE(DATARS_DTH(0));
  nbMvts = 1;
  j=0;

  /* Parcour de tous les points */
  for (i=1; i<DATARS_NB; i++) {

    isCircle = IS_CIRCLE(DATARS_DTH(i));
    stopBefore = IS_STOP(DATARS_DL(j)*DATARS_DL(i));

    /* 2 cercles qui tournent dans le meme sens */
    if (!stopBefore && isCircle && prevIsCircle 
	&& DATARS_DTH(j)*DATARS_DTH(i) > 0.) {

      /* Rayons differents -> on ne peut les fusionner */
      if(fabs(DATARS_DL(j)/DATARS_DTH(j) - DATARS_DL(i)/DATARS_DTH(i))
	 > 0.001) {
	printf ("piloRSFiltre error: 2 consecutive rotations (%d and %d) in same direction\n"
		"                    with different radius (not treated yet, sorry)\n", i-1, i);
	 return(S_pilo_bad_parameters);
      }

      /* Fusion des 2 cercles */
      printf ("piloRSFiltre warning: 2 consecutive rotations in same direction\n");
      DATARS_DTH(j)+=DATARS_DTH(i);
      DATARS_DL(j)+=DATARS_DL(i);
      continue;
    }
    /* Segment suivant */
    else {
      j++;
      if(j != i) {
	DATARS_DTH(j) = DATARS_DTH(i);
	DATARS_DL(j) = DATARS_DL(i);
      }	
    }
    /* Lissage avec le mouvement precedent */
    if (!stopBefore) {
      
      /* Lissage cercle-cercle */
      if (isCircle && prevIsCircle) {
	
	/* Clotho + segment */
	nbMvts += 2;
      }

      /* Clotho */
      nbMvts++;
    }
    /* Le mouvement circle or segment */
    nbMvts++;
    prevIsCircle = isCircle;
  }    /* for */
  
  /* gotoEnd */
  nbMvts++;
  DATARS_NB = j+1;
  *nMvts = nbMvts;
  return OK;
}


/********************************************************************
 *
 *  piloRSControlResult
 *
 *   return: bilan
 */
int piloRSControlResult(int nbMvts)
{
  double xRef, yRef, thRef;
  double dth;
  int i;
  
  /* Compute final pos according to given data */
  xRef = X_START(0);
  yRef = Y_START(0);
  thRef = THETA_START(0);
  for (i=0; i<DATARS_NB; i++) {
    if (DATARS_DTH(i) == 0.) {
      xRef += DATARS_DL(i) * cos(thRef);
      yRef += DATARS_DL(i) * sin(thRef);
    }
    else {
      dth = DATARS_DTH(i);
      xRef += DATARS_DL(i)/DATARS_DTH(i) * (sin(thRef+dth) - sin(thRef));
      yRef += DATARS_DL(i)/DATARS_DTH(i) * (cos(thRef) - cos(thRef+dth));
      thRef += dth;
    }
  }
  
  /* Compare to computed final pos */
  if(fabs(xRef-X_START(nbMvts-1)) > 0.01 || 
     fabs(yRef-Y_START(nbMvts-1)) > 0.01 ||
     fabs(angleLimit(thRef-THETA_START(nbMvts-1))) > 0.01) {
    printf ("piloRSControlResult error: init: %g %g %g\n"
	    "                   end   wanted: %g %g %g\n"
	    "                   end computed: %g %g %g\n",
	    DATARS_X0, DATARS_Y0, DATARS_THETA0,
	    xRef, yRef, thRef, 
	    X_START(nbMvts-1), Y_START(nbMvts-1), THETA_START(nbMvts-1));
    return S_pilo_ERROR_IN_RS_COMPUTATION;
  }
  return OK;
}


/******************************************************************************
 *
 *  piloRSTypes  -  Determination de la geometrie de la trajectoire.
 *
 *  Description: Enregistre dans la structure smoothStrId:
 *       . les types des mouvements elementaires successifs 
 *         (TURN, SEGMENT, CLOTHO) -voir commentaire piloRSFiltre- 
 *       . les longueurs curviligne et angulaire pour les droites et 
 *         cercles de la courbe de Reed&Shep modifie'e.
 *         (Lissage avec clotho effectue apres par piloRSGeo).
 *
 *  Modif de la courbe de Reed&Shep fourni: 
 *    piloRSGeo ne sait lisser que les enchainements droite/cercle. 
 *    Les enchainements cercle/cercle sans rebroussement sont donc substitues
 *    par des enchainements cercle/droite/cercle. Pour cela les cercles
 *    sont reduits et maintenus tangent aux cercles initiaux 
 *    (en debut, milieu ou fin de l'arc selon qu'ils sont precedes ou suivit 
 *    d'un segment ou d'un cercle). La nouvelle droite est la tangente 
 *    commune a ces 2 nouveaux cercles. 
 *
 *  Retourne:  OK ou bilan
 */


int piloRSTypes (int *nbMvts)

{
  int numMvt;
  int i;
  BOOL stopBefore, stopAfter;
  BOOL isCircle, prevIsCircle, nextIsCircle;
  PILO_CIRCLE_TYPE circleType, prevCircleType;
  double lV, dthV;
  double ro, dl, dth;
  int nbMvtsReal;

  double fact;
  double poid=3.;
  PILO_CIRCLE_TYPE  prevPrevCircleType;
  double prevDthV=0.;
  double newPrevDthV, prevlV;
  
  i=0;
  stopBefore = TRUE;
  prevIsCircle = FALSE;
  isCircle = IS_CIRCLE(DATARS_DTH(i));
  numMvt = -1;
  circleType = CIRCLE_START;
  prevCircleType = CIRCLE_START;
  prevPrevCircleType = CIRCLE_START;

  /* Determiner les types de mouvement */
  for (i=0; i<DATARS_NB; i++) {

    /* New state */
    if (i != DATARS_NB-1) {
      stopAfter = IS_STOP(DATARS_DL(i)*DATARS_DL(i+1));
      nextIsCircle = IS_CIRCLE(DATARS_DTH(i+1));
    }
    else {
      stopAfter = TRUE;
      nextIsCircle = FALSE;
    }
    dl = DATARS_DL(i);
    dth = DATARS_DTH(i);
    dthV = 0.;
        
    /* circle */
    if (isCircle) {
      ro = dl/dth;
      circleType = CIRCLE_START;
      if ((prevIsCircle && !stopBefore) || (nextIsCircle && !stopAfter)) {
	ro *= PILO_REDUCE_RO;
	dl = ro*dth;
      }
    }
    else
      ro = infinity();

    /*  Smooth with previous motion */
    if (!stopBefore) {
      
      /* Smooth circle-circle */
      if (isCircle && prevIsCircle) {

	if (nextIsCircle && !stopAfter) circleType = CIRCLE_CENTER;
	else circleType = CIRCLE_END;

	/* Compute common tangent */
	fact = 1.;
	while (!computeTangentWithPrev(i, prevCircleType, circleType,
				       fact*RO_START(numMvt), fact*ro,
				       &dthV, &lV)) {
	  fact = (fact*poid + 1/PILO_REDUCE_RO)/(poid + 1);
	  printf ("piloRSTypes: increase radius -> %g circles %d and %d\n",
		  fact*ro, i-1, i);
	}

	/* Increase radius of circles */
	newPrevDthV = prevDthV;
	if (fact != 1.) {
	  ro *= fact;
	  RO_START(numMvt) *= fact;

	  /* Back-track new radius of prev circle */
	  if(prevCircleType == CIRCLE_CENTER) {
	    if (!computeTangentWithPrev(i-1, 
					prevPrevCircleType, prevCircleType, 
					RO_START(numMvt-4), RO_START(numMvt),
					&newPrevDthV, &prevlV)) {
		printf ("piloRSTypes: rien ne va plus !!!\n");
		return S_pilo_ERROR_IN_RS_COMPUTATION;
	      }
	      else {
		/* Change prev prev seg */
		VAR_LIN(numMvt-2) = prevlV;
		/* Change prev prev circle */
		VAR_ANG(numMvt-4) += newPrevDthV - prevDthV;
		VAR_LIN(numMvt-4) = RO_START(numMvt-4) * VAR_ANG(numMvt-4);
	      }
	  }
	} 

	/* Modif previous circle (the turn ends sooner) */
	VAR_ANG(numMvt) += dthV - (newPrevDthV - prevDthV);
	VAR_LIN(numMvt) = RO_START(numMvt) * VAR_ANG(numMvt);
	
	/* Add clotho and segment */
	TYPE_MVT(++numMvt) = CLOTHO;
	SEG_NUM(numMvt) = i;

	TYPE_MVT(++numMvt) = SEGMENT;
	VAR_ANG(numMvt) = 0.;
	VAR_LIN(numMvt) = lV;
	SEG_NUM(numMvt) = i;
	RO_START(numMvt) = infinity();
	C_START(numMvt) = 0.;
    
	/* Modif current circle */
	dth -= dthV;
	dl = ro*dth;
      }

      /* Add clotho */
      TYPE_MVT(++numMvt) = CLOTHO;
      SEG_NUM(numMvt) = i;
    }

    /* Record the current motion (seg or circle) */
    TYPE_MVT(++numMvt) = isCircle ? TURN : SEGMENT;
    VAR_ANG(numMvt) = dth;
    VAR_LIN(numMvt) = dl;
    RO_START(numMvt) = ro;
    if (isCircle) {
      if (ro == 0.) C_START(numMvt) = infinity();
      else C_START(numMvt) = 1./ro;
    }
    else
      C_START(numMvt) = 0.;
    SEG_NUM(numMvt) = i;

    /* Memorize previous state */
    stopBefore = stopAfter;
    prevIsCircle = isCircle;
    isCircle = nextIsCircle;
    prevPrevCircleType = prevCircleType;
    prevCircleType = circleType;
    prevDthV = dthV;
  }

  /* GOTO END */
  TYPE_MVT(++numMvt) = GOTO_END;
  SEG_NUM(numMvt) = i;
  nbMvtsReal = numMvt+1;
/*  printf ("piloRSCharacterize: %d mvts   -> %d mvts\n", *nbMvts, nbMvtsReal);*/
  if (nbMvtsReal > *nbMvts) {
    printf ("piloRSCharacterize: incoherence nombre mouvement elementaires\n");
    return S_pilo_ERROR_IN_RS_COMPUTATION;
  }
  *nbMvts = nbMvtsReal;

  return OK;
}

/******************************************************************************
 *
 *  piloRSGeo  -  Determination de la geometrie de la trajectoire.
 *
 *  Description: Calcul des clothos qui lissent les enchainements 
 *               droite/cercle (sans rebrousement); ainsi que les longueurs 
 *               restant a executer sur les segments et arcs de cercle.
 *               N'utilise que la structure piloSmoothTrajId.
 *               Ne lisse pas les enchainements cercle/cercle.
 *
 *  Retourne:  OK ou bilan
 */


int piloRSGeo (int *nbMvts)

{
  int i;
  PILO_TYPE_MVT typeMvt, nextTypeMvt;
  BOOL prevIsClotho, nextIsClotho;
  double roVirage, thetaV, thetaJMax, lC;
  double dlVirage, prevDlVirage, dthVirage, kCloth;
  double pVCloth, xCloth=0., yCloth=0., thCloth, distCloth=0.;
  double thRot, thRotEnd;
  double thMove;
  int bilan;

  /* Initialisations */
  typeMvt = TYPE_MVT(0);
  prevIsClotho = FALSE;
  prevDlVirage = 0.;

  /*
   * Compute the geometry of each motion
   */
  for (i=0; i<*nbMvts-1; i++) {
    
    nextTypeMvt = TYPE_MVT(i+1);
    nextIsClotho = nextTypeMvt == CLOTHO ? TRUE : FALSE;

    switch(typeMvt) {

      /* Clotho */
    case CLOTHO:

      prevIsClotho = TRUE;
      /* Report computation during smoothing */
      dlVirage = prevDlVirage;
      break;

      /* Segment */
    case SEGMENT:

      /* Next is cusp: seg ends here */
      if (!nextIsClotho) {
	VAR_LIN(i) -= prevDlVirage;
	/* End of move */
	thMove = THETA_START(i);
	THETA_START(i+1) = thMove;
	X_START(i+1) = X_START(i) + VAR_LIN(i) * cos(thMove);
	Y_START(i+1) = Y_START(i) + VAR_LIN(i) * sin(thMove);
	dlVirage = 0.;
      }
      /* Report computation during smoothing */
      else dlVirage = prevDlVirage;

      prevIsClotho = FALSE;
      break;

      /* Circle: computation of the smoothing if any */
    case TURN:

      roVirage = RO_START(i);
      dlVirage = 0.;

      /* Smooth of the circle */
      if (prevIsClotho || nextIsClotho) {

	/* Lengths of the turn to be smooth */
	thetaV = VAR_ANG(i);
	thetaJMax = thetaV * PILO_REDUCE_ANGLE;
	if (prevIsClotho && nextIsClotho) {
	  thetaV /= 2.;
	  thetaJMax /= 2.;
	  lC = SIGN(VAR_LIN(i-2)) * PILO_REDUCE_LENGTH
	       * MIN(fabs(VAR_LIN(i-2)), fabs(VAR_LIN(i+2)));
	}
	else if (prevIsClotho) lC = VAR_LIN(i-2) * PILO_REDUCE_LENGTH;
	else lC = VAR_LIN(i+2) * PILO_REDUCE_LENGTH;

	/* Smooth */
	if ((bilan = computeSmoothLineCircle(lC, thetaV, thetaJMax, RO_START(i),
					     &dlVirage, &dthVirage, &roVirage, 
					     &kCloth)) != OK) {
	  printf ("piloRSGeo: computeSmoothLineCircle failed, motion %d (circle).\n", i);
	  return bilan; 
	}
	pVCloth = sqrt (2.*fabs(dthVirage) / PI );
	xCloth = kCloth * integCosFresnel(pVCloth);
	yCloth = SIGN(dthVirage) * kCloth * integSinFresnel (pVCloth);
	distCloth = SIGN(kCloth) * kCloth * kCloth / (PI * fabs(roVirage));

	/* New radius of circle */
	RO_START(i) = roVirage;
	C_START(i) = 1/roVirage;

	/* Record previous motions */
	if (prevIsClotho) {

	  /* Reduce angle variation on circle */
	  VAR_ANG(i) -= dthVirage;
  
	  /* End of previous move */
	  VAR_LIN(i-2) -= prevDlVirage + dlVirage;
	  thCloth = THETA_START(i-2);
	  THETA_START(i-1) = thCloth;
	  X_START(i-1) = X_START(i-2) + VAR_LIN(i-2) * cos(thCloth);
	  Y_START(i-1) = Y_START(i-2) + VAR_LIN(i-2) * sin(thCloth);

	  /* Previous clotho */
	  K_TRAJ(i-1) = kCloth;
	  VAR_ANG (i-1) = dthVirage;
	  RO_START(i-1) = infinity();
	  C_START(i-1) = 0.;
	  VAR_LIN (i-1) = distCloth;

	  /* End of previous clotho */
	  X_START(i) = X_START(i-1)
		       + xCloth * cos(thCloth) - yCloth * sin(thCloth);
	  Y_START(i) = Y_START(i-1)
		       + xCloth * sin(thCloth) + yCloth * cos(thCloth);
	  THETA_START(i) = thCloth + dthVirage;
	} /* Smooth before */

	/* Reduce angle variation on circle */
	if (nextIsClotho) VAR_ANG(i) -= dthVirage;

      } /* Smooth circle */

      /* End of circle */
      thRot = THETA_START(i);
      thRotEnd = thRot + VAR_ANG(i);
      THETA_START(i+1) = thRotEnd;
      VAR_LIN(i) = roVirage * VAR_ANG(i);
      X_START(i+1) = X_START(i) + roVirage * (sin(thRotEnd) - sin(thRot));
      Y_START(i+1) = Y_START(i) + roVirage * (cos(thRot) - cos(thRotEnd));

      /* Next clotho */
      if (nextIsClotho) {
	VAR_ANG(i+1)  = dthVirage;
	RO_START(i+1) = roVirage;
	C_START(i+1)  = 1./roVirage;
	K_TRAJ(i+1)   = kCloth;
	VAR_LIN(i+1) = distCloth;
	
	/* End of next clotho */
	THETA_START(i+2) = THETA_START(i+1) + dthVirage;
	thCloth = THETA_START(i+2);
	X_START(i+2) = X_START(i+1)
		     + xCloth * cos(thCloth) + yCloth * sin(thCloth);
	Y_START(i+2) = Y_START(i+1)
		     + xCloth * sin(thCloth) - yCloth * cos(thCloth);
      }
      prevIsClotho = FALSE;
      break;

    default:
      printf ("piloRSGeo: non mais qu'est-ce que tu fout la` ?!? (typeMvt %d)\n",
	      typeMvt);
      return S_pilo_ERROR_IN_RS_COMPUTATION;
    }  /* switch */
    
    /* Record length reduction on segment */
    prevDlVirage = dlVirage;

    /* Record previous motion type */
    typeMvt = nextTypeMvt;

  } /* For */

  /* GOTO_END */
  if (TYPE_MVT(i) != GOTO_END) {
    printf ("piloRSGeo: unvalid final type\n");
    return S_pilo_ERROR_IN_RS_COMPUTATION;
  }
  VAR_LIN(i) = 0.;
  VAR_ANG(i) = 0.;
  *nbMvts = i+1;
  return OK;
}


/********************************************************************
 *
 *  computeTangentWithPrev - Remplace 2 cercles tangents par 2 cercles
 *                          disjoints - de rayon plus petit - lies par une 
 *                          tangente commune.
 *
 *  Les cercles reduits sont tangents aux cercles initiaux. Et ce soit
 *  au debut du virage (CIRCLE_START), soit au milieux du virage 
 *  (CIRCLE_CENTER) ou a la fin du virage (CIRCLE_END). 
 *
 *  Retourne: . length: longueur du segment tangent
 *            . dTheta: coord polaire point de tangence sur 1er cercle,
 *                      i.e. nouvelle variation angulaire sur 1er cercle.
 */
static BOOL computeTangentWithPrev(int numData, 
				   PILO_CIRCLE_TYPE prevCircleType, 
				   PILO_CIRCLE_TYPE circleType, 
				   double newRo1, double newRo2,
				   double *dTheta, double *length)
{
  CIRCLE C1, C2;         /* Cercles initiaux */
  CIRCLE c1, c2;         /* Cercles reduits */
  double dt1, dt2;       /* Variations angulaires sur les cercles initiaux */
/*  double phi;*/

  /* 
   * Reduction 1er cercle (origine des angles au point de tangence)
   */
  C1.x = 0.;
  C1.y = 0.;
  dt1 = DATARS_DTH(numData-1);
  C1.r = fabs(DATARS_DL(numData-1))/dt1;

  /* Tangence en milieu de virage */
  if (prevCircleType == CIRCLE_CENTER) dt1 /= 2.;

  c1 = tangentCircle(&C1, -dt1, SIGN(C1.r)*fabs(newRo1));

  /* 
   * Reduction 2nd cercle 
   */
  dt2 =  DATARS_DTH(numData);
  C2.r = fabs(DATARS_DL(numData))/dt2;
  C2.x = fabs(C1.r - C2.r);
  C2.y = 0.;

  /* Tangence en milieu de virage */
  if (circleType == CIRCLE_CENTER) dt2 /= 2.;

  c2 = tangentCircle(&C2, dt2+PI, SIGN(C2.r)*fabs(newRo2));

  /* 
   * Tangent between the circles 
   */
  if (tangentSegBetweenCircles(&c1, &c2, length, dTheta) != OK)
    return FALSE;

  *length *= SIGN(DATARS_DL(numData));

  /* 
   * Test validite
   */
  /* La tangente ne demarre pas dans le secteur angulaire du 1er cercle */
  if (*dTheta*dt1 > 0.)
    printf ("computeTangentWithPrev: warning: the tangent segment starts too late\n"
	    "                       (dTheta %g, dt1 %g)\n", *dTheta, dt1);
  else
    if (fabs(*dTheta) > fabs(dt1)) {
      printf ("computeTangentWithPrev: warning, the tangent segment starts "
	      "too soon:\n"
	      "                        |dTheta|=|%g| > |dt1|=|%g|\n", 
	      *dTheta, dt1);
      return FALSE;
    }
  
  /* La tangente ne termine pas dans le secteur angulaire du 2nd cercle */
  if (*dTheta*dt2 < 0.)
    printf ("computeTangentWithPrev: warning: the tangent segment ends too soon\n"
	    "                       (dTheta %g, dt2 %g)\n", *dTheta, dt2);
  else 
    if (fabs(*dTheta) > fabs(dt2)) {
      printf ("computeTangentWithPrev: warning, the tangent segment ends "
	      "too late:\n"
	      "                        |dTheta|=|%g| > |dt2|=|%g|\n", 
	      *dTheta, dt2);
      return FALSE;
    }

  return TRUE;
}

/********************************************************************
 *
 *  tangentCircle   -  Determine le cercle tangent a un autre cercle.
 *
 *                     factRadius: raport rayon du nouveau cercle.
 *                     theta: coordonnee polaire du point de tangence.
 *
 */
static CIRCLE tangentCircle(CIRCLE *circle, double theta, double newRo)
{
  CIRCLE newCircle;

  if (circle->r<0.) theta+=PI;
  newCircle.r = newRo;
  newCircle.x = circle->x + (circle->r - newCircle.r) * cos(theta);
  newCircle.y = circle->y + (circle->r - newCircle.r) * sin(theta);
  
  return newCircle;
}


/********************************************************************
 *
 *  tangentSegBetweenCircles   -  Determine le segment tangent a 2 cercles.
 *
 *  ATTENTION: le signe du rayon indique le sens de rotation sur le cercle.
 * 
 *  Choix de la tangente: 
 *     La tangente doit respecter les sens de rotation indiques par le signe
 *     des rayons (R>0: sens trigo). (Imaginer une courroie qui part du  
 *     1er cercle et qui est entrainee par le 2nd.)
 *     -> si R1.R2 > 0 alors tangente inferieure (r1>0) ou superieure  (r1<0);
 *        sinon             diagonale ascendante (r1>0) ou descendante (r1<0).
 *
 *  Retourne: length: longueur du segment ( >= 0 )
 *            theta:  coord polaire point de tangence au 1er cercle [-Pi, Pi]
 *                    (direction du segment = theta + SIGN(c1.r) * Pi)
 *
 *  Return: ERROR si les cercles ne sont pas disjoints.
 */
static STATUS tangentSegBetweenCircles(CIRCLE *c1, CIRCLE *c2, 
				       double *length, double *theta) 
  
{
  double dx, dy;          /* Vecteur entre els centres */
  double d;               /* Distance entre les centres */
  double alpha;           /* Direction de la droite passant par les centres */
  double rap;      

  dx = c2->x-c1->x;
  dy = c2->y-c1->y;

  d = sqrt(SQR(dx) + SQR(dy));

  rap = (c1->r-c2->r)/d;

  /* Cercles non disjoints */
  if (fabs(rap) > 1.) {
    if(c1->r*c2->r > 0.)
      printf ("tangentSegBetweenCircles: cercles de signes opposes non-disjoints\n");
    else
      printf ("tangentSegBetweenCircles: un cercle englobe l'autre\n");
    return ERROR;
  }
  alpha = atan2(dy,dx);
  *theta = - SIGN(c1->r) * acos(rap*SIGN(c1->r)) + alpha;

/*  *phi = asin(rap);*/

  *length = sqrt(SQR(d) - SQR(c1->r-c2->r));

  return OK;
}


/**************************************************************************
 *
 *  computeSmoothLineCircle - Lisse enchainement droite-cercle par clotho
 *
 *  Description: le cercle initial doit etre tangent au segment.
 *   Pour faire l'enchainement droite-clotho-cercle, le cercle va etre reduit.
 *   Il est maintenu tangent au cercle initial au point indique (par phiTan).
 *
 * Entree:
 *  lMax  : Longueur max du segment utilisable pour le virage
 *  phiTan: Point de tangence pour le petit cercle
 *  phiMax: Angle max utilisable pour le virage
 *  ro    : Rayon du cercle initial
 * Sortie:
 *  lV     : Longueur de segment utilisee
 *  phiV   : Angle robot a la jonction cercle/clotho
 *  rV     : Rayon du cercle reduit
 *  kClotho: Constante de la clotho
 *
 *  return: OK ou bilan
 */

static int computeSmoothLineCircle(double lMax, double phiTan, 
				   double phiMax, double ro,
				   double *lV, double *phiV,
				   double *rV, double *kClotho)
{
  double x0;    /* Ordonnee cercle dans repere clotho, ie longueur virage */
  double factR;  /* nouveau rayon du cercle */
  double phiJ;  /* Angle de jonction sur petit cercle */
  double sqrt2PhiJ;
  double fabsPhiTan, cosPhiTan, sinPhiTan;
  double signL, signPhiTan, signPhiMax, signRo, fabsRo;

  fabsPhiTan = fabs(phiTan);
  fabsRo = fabs(ro);

  signL = SIGN(lMax);
  signPhiTan = SIGN(phiTan);
  signPhiMax = SIGN(phiMax);
  signRo = SIGN(ro);

  /* Test validite */
  if (fabs(phiMax) > fabsPhiTan){
    printf ("computeSmoothLineCircle: incoherence: phiMax > phiTan\n");
    return S_pilo_bad_parameters;
  }
  if(signPhiTan*signPhiMax < 0.){
    printf ("computeSmoothLineCircle: incoherence: phiTan * phiMax < 0\n");
    return S_pilo_bad_parameters;
  }
  if(signRo * signPhiTan * signL < 0.) {
    printf ("computeSmoothLineCircle: incoherence: ro * phiTan * l < 0\n");
    return S_pilo_bad_parameters;
  }

  cosPhiTan = cos(fabsPhiTan);
  sinPhiTan = sin(fabsPhiTan);

  /* Angle de jonction souhaite */
  phiJ = MIN(fabs(phiMax), PHI_WANTED);
  phiJ /= PILO_REDUCED_PHI_J;

  /* Calcul du rayon et de la longeur de la clotho */
  do {
    phiJ *= PILO_REDUCED_PHI_J;

    sqrt2PhiJ = sqrt(2*phiJ);
    
    factR = (1. - cosPhiTan)
	      / (SQRT_PI * sqrt2PhiJ * integSinFresnel(sqrt2PhiJ/SQRT_PI) 
		 + cos(phiJ) - cosPhiTan);
    
    x0 = fabsRo * 
	 (factR * (SQRT_PI*sqrt2PhiJ*integCosFresnel(sqrt2PhiJ/SQRT_PI)
		   - sin(phiJ))
	  - (1. - factR) * sinPhiTan);
    
  } while (x0 > fabs(lMax));

  *lV   = signL * x0;
  *phiV = signPhiTan * phiJ;
  *rV   = ro * factR;
  *kClotho = signL * fabsRo * factR * SQRT_PI * sqrt2PhiJ;

/*  printf ("Cloth: phiV %g l %g phi %g r %g k %g\n", phiTan, *lV, *phiV, signRo*factR, *kClotho);*/

  return OK;
}
