/*- -*- Mode: C -*- ----------------------------------------------------
 *    
 *                         -- C N R S -- 
 *         Laboratoire d'Automatique et d'Analyse des Systemes 
 *                 7 Avenue du colonel Roche   
 *                     31 077 Toulouse Cedex
 *
 *  Auteur           : Sara Fleury
 *  Groupe           : Robotique et Intelligence Artificielle
 *  Nom du fichier   : piloSpeedLib.c
 *  Date de creation : 1994
 * 
 *  Description      : Calcul de profiles de vitesses pour des trajectoires
 *                     composees de segments, arcs de cercle, clothoides
 *                     et anti-clothoides.
 *
 *  Copyright (C) 1994 LAAS/CNRS.
 *----------------------------------------------------------------------*/
#ifdef UNIX
#  include <stdio.h>
#  include <portLib.h>

/* #  include <sunmath.h> */
#  include <math.h>

#else
#  include <vxWorks.h>
#endif
#include <stdio.h>
#include <math.h>

#include "h2mathLib.h"

#include "codels/piloCmdTaskCodels.h"
#include "server/piloError.h"
#include "piloStruct.h"
#include "piloConst.h"
#include "piloSpeedLibProto.h"

#define MIN3(x,y,z) (MIN((x),MIN((y),(z))))
#define MAX3(x,y,z) (MAX((x),MAX((y),(z))))


static void reduceSpeedBackward(int i, double reduceSpeed);

/*****************************************************************************
 *
 *  piloSpeedProfiles  -  Determination des vitesses sur la trajectoire
 *
 *  Hyp: 1. CLOTHO a` v constante, ANTI a` w constante
 *       2. Pas de discontinuite de tangente
 *       3. Stop ne s'enchaine qu'avec un SEGMENT ou un TURN
 *       4. Pas de discontinute de courbure en dehors du stop
 *          (donc en dehors d'enchainement SEGMENT-TURN ou TURN-TURN)
 *
 *  Description: 
 *  1. Calcul des vitesses max initiales et finales pour chaque type de segment. 
 *  2. Test de transition: 
 *    Detection: vitesses max initiales < vitesses finales mouvement precedent 
 *    2 cas possibles: 1. Limitation en accel sur CLOTHO ou ANTI 
 *                     2. Limitation en accel sur TURN ou SEG suivit d'un STOP
 *    Remede: on reduits les vitesses des mouvements precedents jusqu'a ce qu'on 
 *            retrouve un mouvement de type TURN ou SEG.
 *    Limites de cette solution: pas de limitation en deceleration lors du 
 *                               backtrack.
 *
 *  Retourne:  OK ou bilan
 */


int piloSpeedProfiles (int nbMvts)

{
  int i;
  double reduceSpeed;
  BOOL nextIsStop;

/*   V_START(0) = 0.; */
/*   W_START(0) = 0.; */

  for (i=0; i<nbMvts-1; i++) {
    
    switch(TYPE_MVT(i)) {
      /* CLOTHO   (v = cst) */
    case CLOTHO:
      reduceSpeed = computeClothoProfiles(i);
      if (reduceSpeed < 0.9999)
	reduceSpeedBackward(i, reduceSpeed);
      break;
      
      /* ANTI   (w = cst) */
    case ANTI_CLOTHO:
      reduceSpeed = computeAntiProfiles(i);
      if (reduceSpeed < 0.9999)
	reduceSpeedBackward(i, reduceSpeed);
      break;
      
      /* SEGMENT */
    case SEGMENT:
      nextIsStop = (VAR_LIN(i)*VAR_LIN(i+1) < 0. || K_TRAJ(i+1) == 0.) ? 
		   TRUE : FALSE;
      if(nextIsStop) {
	V_START(i+1) = 0.;
	computeSegProfiles(i, TRUE);
      }
      else {
	V_START(i+1) = VMAX_REF;
	computeSegProfiles(i, FALSE);
      }
      break;
      
      /* TURN */
    case TURN:
      nextIsStop = (VAR_LIN(i)*VAR_LIN(i+1) < 0. || K_TRAJ(i+1) == 0.) ? 
		   TRUE : FALSE;
      if(nextIsStop) {
	W_START(i+1) = 0.;
	computeTurnProfiles(i, TRUE);
      }
      else {
	W_START(i+1) = WMAX_REF;
	computeTurnProfiles(i, FALSE);
      }
      break;

      /* ERROR */
    case GOTO_END:
      printf ("piloRS: GOTO_END is not last motion!\n");
      return S_pilo_ERROR_IN_RS_COMPUTATION;

    default:
      printf ("piloRS: unknown type movement\n");
      return S_pilo_ERROR_IN_RS_COMPUTATION;
      
    } /* switch */

  } /* for */

  /* On est arrive' */
  if (TYPE_MVT(i) != GOTO_END) {
    printf ("piloRS: last motion is not GOTO_END!\n");
    return S_pilo_ERROR_IN_RS_COMPUTATION;
  }
  V_START(i) = 0.;
  W_START(i) = 0.;

  return OK;
}

/*****************************************************************************
 *
 *  computeSegProfiles  - Calcul des profiles de vitesse pour un segment
 *
 * Description: A partir de la description du segment dans la structure
 *   dont l'indice i est passe' en parametre, calcul les profiles de vitesse.
 *   (ie vitesses depart/arrivee, vitesses max, accelerations, 
 *   temps d'execution). Ces valeurs sont majorees par VMAX_REF, 
 *   WMAX_REF, ALIN_REF, AANG_REF.
 *
 *   La vitesse de depart souhaitee est V_START(i), et celle d'arrivee 
 *   V_START(i+1). Selon le flag initSpeedFixed, on fixe la vitesse de depart
 *   celle d'arrivee etant prise comme seuil max, ou les vitesses de depart 
 *   ET d'arrivee. Par exemple, si l'on souhaite s'arreter en bout de 
 *   segment, initialiser V_START(i+1) a` zero, sinon mettre la valeur max. 
 *
 *   La fonction retourne le rapport entre la valeur souhaitee de
 *   la vitesse d'arrivee et la valeur calculee (ie vitesse max 
 *   atteignable en respectant les contrainte en vitesses et acceleration).
 *   Si la vitesse de consigne etait nule, alors elle retourne 1 si la 
 *   vitesse calculee est nule ou la norme de cette vitesse en cas contraire.
 */
void computeSegProfiles(int i, BOOL bothSpeedsFixed)
{
  double vStart, wStart;         /* Vitesses depart (valeur absolue) */
  double vEnd, wEnd;             /* Vitesses arrive (valeur absolue) */
  double vS2, vE2;               /* Carre' des vitesses depart/arrive */
  double accelLin;               /* valeur absolue acceleration */
  double varLin;                 /* valeur absolue variations */
  double deltaV2Max;             /* 2.aMax.d = vf^2 - vs^2 (valeur absolue) */
  double forward;                /* Signe du mouvement */
  double vMax;

  /* Init variables */
  varLin = fabs(VAR_LIN(i));
  deltaV2Max = 2*ALIN_REF*varLin;
  forward = SIGN(VAR_LIN(i));
  
  /* Limitation des vitesses */
  vStart = MIN(fabs(V_START(i)), VMAX_REF);
  vS2 = SQR(vStart);

  /* Vitesse initiale fixee: calcul de la vitesse finale */
  if(!bothSpeedsFixed) {
    /* Limitation en vitesse et accel lineaire */
    vEnd = MIN(VMAX_REF, sqrt(vS2 + deltaV2Max));
    vE2 = SQR(vEnd);
  }

  /* Vitesses initiale et finale fixees */
  else {
    vEnd = MIN(fabs(V_START(i+1)), VMAX_REF);
    vE2 = SQR(vEnd);
    /* Limitation en acceleration */
#ifdef VERBOSE
    if (fabs(vE2-vS2) > deltaV2Max)
      printf ("PILO speed warning: accel. lin. augmentee (seg %d)\n", i);
#endif
  }
  
  /* Vitesses angulaires */
  wStart = 0.;
  wEnd = 0.;
  
  /* On cherche a diminuer l'acceleration lineaire */
  accelLin = MAX(fabs(vE2 - vS2)/(2.*varLin),
		 ALIN_REF/PILO_REDUC_ACCEL_LIN);
  ACCEL_LIN(i) = forward * accelLin;
  ACCEL_ANG(i) = 0.;
  
  /* Accelerations et vitesses max */
  vMax =  MIN(VMAX_REF, sqrt(accelLin * varLin + (vS2+vE2)/2.));
  V_MAX(i) = forward * vMax;
  W_MAX(i) = 0.;
  
  /* Enregistre les vitesses de depart */
  V_START(i) = forward * vStart;
  W_START(i) = 0.;
  TIME_MIN(i) = ((SQR(vMax - vStart) + SQR(vMax - vEnd))/(2.*accelLin) 
		 + varLin) /vMax;
  
  /* Vitesse d'arrivee */
  V_START(i+1) = forward * vEnd;
  W_START(i+1) = 0.;
}


/*****************************************************************************
 *
 *  computeTurnProfiles  - Calcul des profiles de vitesse pour un arc de 
 *                         cercle.
 *
 *  Idem que computeSegProfiles pour des arcs de cercles (remplacer les
 *  vitesses lineaires par des vitesses angulaires)
 */

void computeTurnProfiles(int i, BOOL bothSpeedsFixed)
{
  double vStart, wStart;          /* Vitesses depart (valeur absolue) */
  double vEnd, wEnd;              /* Vitesses arrive (valeur absolue) */
  double vS2, vE2, wS2, wE2;      /* Carre' des vitesses depart/arrive */
  double accelAng;                /* valeur absolue acceleration */
  double varAng, varLin;          /* valeur absolue variations */
  double deltaV2Max, deltaW2Max;  /* 2.aMax.d = vf^2 - vs^2 (valeur absolue) */
  double accelAngMin, accelLinMin; /* Val. abs. des accel. min. */
  double roStart;                  /* Val. abs. du rayon */
  double wMax;
  double forward, trigo;

  /* Initialisation */
  roStart = fabs(RO_START(i));
  varAng = fabs(VAR_ANG(i));
  varLin = fabs(VAR_LIN(i));
  deltaW2Max = 2*AANG_REF*varAng;
  deltaV2Max = 2*ALIN_REF*varLin;
  forward = SIGN(VAR_LIN(i));
  trigo = SIGN(VAR_ANG(i));

  /* Limitation des vitesses de depart */
  wStart = MIN(fabs(W_START(i)), WMAX_REF);
  if(roStart ==  0.) vStart = 0.;
  else {
    vStart = wStart*roStart;
    if(vStart > VMAX_REF) {
      vStart = VMAX_REF;
      wStart = vStart/roStart;
    }
  }
  vS2 = SQR(vStart);
  wS2 = SQR(wStart);
  
  /* Vitesses initiales fixees: calcul de la vitesse finale */
  if(!bothSpeedsFixed) {

    /* Limitation en vitesse et accel ang */
    wEnd = MIN(WMAX_REF, sqrt(wS2 + deltaW2Max));

    /* Limitation en vitesse et accel lin */
    if(roStart == 0.) vEnd = 0.;
    else {
      vEnd = MIN3(VMAX_REF, wEnd * roStart, sqrt(vS2 + deltaV2Max));
      wEnd = MIN(wEnd, vEnd/roStart);
    }
    vE2 =  SQR(vEnd);
    wE2 =  SQR(wEnd);
  }

  /* Vitesses initiales et finales fixees */
  else {
    wEnd = MIN(fabs(W_START(i+1)), WMAX_REF);
    vEnd = wEnd * roStart;
    if(vEnd > VMAX_REF) {
      vEnd = VMAX_REF;
      wEnd = vEnd/roStart;
    }
    vE2 =  SQR(vEnd);
    wE2 =  SQR(wEnd);
#ifdef VERBOSE
    if(fabs(vE2-vS2) > deltaV2Max)
      printf("PILO speed warning: accel. lin. augmentee (cercle %d)\n", i);
    if(fabs(wE2-wS2) > deltaW2Max)
      printf("PILO speed warning: accel. ang. augmentee (cercle %d)\n", i);
#endif
  }
  
  /* Acceleration angulaire minimale */
  accelAngMin = fabs(wE2 - wS2)/(2. * varAng);

  /* Accelerations et vitesses max */
  if (roStart == 0.) {
    accelAng = MAX(accelAngMin, AANG_REF/PILO_REDUC_ACCEL_ANG);
    wMax = WMAX_REF;
  }
  else {
    accelAng = MIN(AANG_REF, ALIN_REF/roStart)/PILO_REDUC_ACCEL_ANG;
    accelLinMin = fabs(vE2 - vS2)/(2. * varLin);
    accelAng = MAX3(accelAng, accelAngMin, accelLinMin/roStart);
    wMax = MIN(WMAX_REF, VMAX_REF/roStart);
  }
  ACCEL_ANG(i) = trigo * accelAng;
  W_MAX(i) = trigo * 
	     MIN(wMax, sqrt (accelAng*varAng + (wS2+wE2)/2.));
  V_MAX(i) = W_MAX(i) * RO_START(i);
  ACCEL_LIN(i) = ACCEL_ANG(i) * RO_START(i);
  
  /* Enregistre les vitesses de depart */
  V_START(i) = forward * vStart;
  W_START(i) = trigo * wStart;
  TIME_MIN(i) = ((SQR(wMax - wStart) + SQR(wMax - wEnd))/(2.*accelAng) 
		 + varAng) /wMax;
  
  /* Vitesse d'arrivee */
  V_START(i+1) = forward * vEnd;
  W_START(i+1) = trigo * wEnd;
}


/*****************************************************************************
 *
 *  computeClothoProfiles  - Calcul des profiles de vitesses pour une clotho
 *
 * Description: A partir de la description de la clotho dans le structure
 *   d'indice i passe en parametre, calcul les profiles de vitesse.
 *   (ie vitesses depart/arrivee, vitesses max, acceleration et
 *   temps d'execution).
 *
 *   Les vitesses de depart souhaitees sont les champs V_START et W_START de
 *   cette structure. 
 *   Si ces vitesses sont trop grandes, elles sont modifiees.
 *   La fonction retourne le rapport entre l'ancienne et la nouvelle valeur.
 *   Ce rapport vaut donc 1 si les valeurs initiales peuvent etre respectees.
 *  
 *   Les vitesses finales sont enregistrees dans les champs V_START et W_START
 *   de la structure d'indice i+1.
 */
double computeClothoProfiles(int i)
{
  double cStart, cEnd, cMax;      /* Courbures (valeur absolue) */
  double vStart, wStart;          /* Vitesses depart (valeur absolue) */
  double vEnd, wEnd;              /* Vitesses arrive (valeur absolue) */
  const double sqrtAaOverPi = sqrt(AANG_REF/PI);
  double vStartRef;
  double forward, trigo;

  /* Init variables */
  cStart = fabs(C_START(i));
  cEnd = fabs(C_START(i+1));
  cMax = MAX(cStart, cEnd);
  vStartRef = fabs(V_START(i));
  forward = SIGN(VAR_LIN(i));
  trigo = SIGN(VAR_ANG(i));
  
  /* Limitation vitesses et en acceleration angulaire */
  vStart = MIN3(vStartRef, VMAX_REF, WMAX_REF/cMax);
  vStart = MIN(vStart, fabs(K_TRAJ(i)) * sqrtAaOverPi);
  
  /* Vitesses finales */
  vEnd = vStart;
  wEnd = vEnd * cEnd;
  
  /* Vitesses de demarage */
  wStart = vStart * cStart;
  
  /* Acceleration */
  ACCEL_ANG(i) = SIGN(VAR_ANG(i)*(wEnd - wStart)) * PI * SQR(vStart/K_TRAJ(i));
  ACCEL_LIN(i) = 0.;
  
  /* Enregistre les vitesses de depart */
  V_START(i) = forward * vStart;
  W_START(i) = trigo * wStart;
  V_MAX(i) =  V_START(i);
  W_MAX(i) =  trigo * MAX(wStart, wEnd);
  TIME_MIN(i) = fabs(VAR_LIN(i))/vStart;
  
  /* Vitesse d'arrivee */
  V_START(i+1) = forward * vEnd;
  W_START(i+1) = trigo * wEnd;

  /* Indiquer si changement de vitesse de depart */
  return vStart/vStartRef;
}

/*****************************************************************************
 *
 *  computeAntiProfiles  - Calcul des profiles de vitesses pour une anti-cloth
 *
 *  Description: voir computeClothoProfiles
 */
double computeAntiProfiles(int i)
{
  double wStartRef;
  double roStart, roEnd, roMax;   /* Rayons (valeur absolue) */
  double vStart, wStart;          /* Vitesses depart (valeur absolue) */
  double vEnd, wEnd;              /* Vitesses arrive (valeur absolue) */
  double forward, trigo;

  /* Limitation en vitesse lineaire */
  roStart = fabs(RO_START(i));
  roEnd = fabs(RO_START(i+1));
  roMax = MAX(roStart, roEnd);
  wStartRef = fabs(W_START(i));
  forward = SIGN(VAR_LIN(i));
  trigo = SIGN(VAR_ANG(i));

  /* Limitation en vitesses et en acceleration lineaire */
  wStart = MIN3(wStartRef, WMAX_REF, VMAX_REF/roMax);
  wStart = MIN(wStart, sqrt(ALIN_REF/fabs(K_TRAJ(i))));
  
  /* Vitesses finales */
  wEnd = wStart;
  vEnd = wEnd * roEnd;
  
  /* Vitesses de demarage */
  vStart = wStart * roStart;
  
  /* Acceleration */
  ACCEL_LIN(i) = K_TRAJ(i) * SQR(wStart);
  ACCEL_ANG(i) = 0.;
  
  /* Enregistre les vitesses de depart */
  V_START(i) = forward * vStart;
  W_START(i) = trigo * wStart;
  V_MAX(i) =  forward * MAX(vStart, vEnd);
  W_MAX(i) =  W_START(i);
  TIME_MIN(i) = fabs(VAR_ANG(i))/wStart;
  
  /* Vitesse d'arrivee */
  V_START(i+1) = forward * vEnd;
  W_START(i+1) = trigo * wEnd;

  /* Indiquer si changement de vitesse de depart */
  return wStart/wStartRef;
}

/*****************************************************************************
 *
 *  reduceSpeedBackward - Reduit les vitesses des segments precedents le
 *                        segment d'indice i selon le rapport transmit en 
 *                        parametre.
 *
 *  Cette fonction est appelee lorsqu'on arrive trop rapidement sur une clotho
 *  ou une anti-clotho. Elle permet donc de reduire les vitesses sur les segments 
 *  precedents jusqu'a ce que l'on rejoigne une portion de droite ou d'arc de
 *  cercle sur laquelle on peut specifier les vitesses de depart ET d'arrivee.
 *
 */
static void reduceSpeedBackward(int i, double reduceSpeed)
{
  int j;
  PILO_TYPE_MVT prevTypeMvt;

  /* Reduction de vitesses precedentes si clotho ou anti-clotho */
  j = i-1;
  while (K_TRAJ(j) != 0.) {
    prevTypeMvt = TYPE_MVT(j);
    /* Reduction vitesses anti-clotho */	  
    if (prevTypeMvt == ANTI_CLOTHO) {
#ifdef VERBOSE
      printf ("piloSpeedProfiles: anti  %d : %6.3f -> %6.3f (rd/s)\n",
	      j, W_START(j), W_START(j+1));
#endif
      W_START(j) = W_START(j+1);
      V_START(j) = W_START(j) * RO_START(j);
      W_MAX(j) = W_START(j);
      V_MAX(j) *=  reduceSpeed;
      ACCEL_LIN(j) *= SQR(reduceSpeed);
      TIME_MIN(j) /= reduceSpeed;
    }
    /* Reduction vitesses clotho */	  
    else if (prevTypeMvt == CLOTHO) {
#ifdef VERBOSE
      printf ("piloSpeedProfiles: cloth %d: %6.3f -> %6.3f (m/s)\n",
	      j, V_START(j), V_START(j+1));
#endif
      V_START(j) = V_START(j+1);
      W_START(j) = V_START(j) * C_START(j);
      V_MAX(j) =  V_START(j);
      W_MAX(j) *=  reduceSpeed;
      ACCEL_ANG(j) *= SQR(reduceSpeed);
      TIME_MIN(j) /= reduceSpeed;
    }
    else printf ("piloSpeedProfiles: ERROR\n");
    j--;
  }
  
  /* Recalcul du profile des vitesses du segment ou cercle precedent */
  if(TYPE_MVT(j) == SEGMENT) {
#ifdef VERBOSE
    printf ("piloSpeedProfiles: seg   %d: %6.3f -> %6.3f (m/s)\n",
	    j, V_MAX(j), V_START(j+1));
#endif
    computeSegProfiles(j, TRUE);
  }
  else if(TYPE_MVT(j) == TURN) {
#ifdef VERBOSE
    printf ("piloSpeedProfiles: turn  %d: %6.3f -> %6.3f (rd/s)\n",
	    j, W_MAX(j), W_START(j+1));
#endif
    computeTurnProfiles(j, TRUE);
  }
  else printf ("piloSpeedProfiles: ERROR\n");
}
